# -*- coding:utf-8 -*-

#  ************************** Copyrights and license ***************************
#
# This file is part of gcovr 8.6, a parsing and reporting tool for gcov.
# https://gcovr.com/en/8.6
#
# _____________________________________________________________________________
#
# Copyright (c) 2013-2026 the gcovr authors
# Copyright (c) 2013 Sandia Corporation.
# Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
# the U.S. Government retains certain rights in this software.
#
# This software is distributed under the 3-clause BSD License.
# For more information, see the README.rst file.
#
# ****************************************************************************

"""
Handle parsing of the json ``.gcov.json.gz`` file format.

Other modules should only use the following items:
`parse_coverage()`

The behavior of this parser was informed by the following sources:

* the *Invoking Gcov* section in the GCC manual (version 11)
  <https://gcc.gnu.org/onlinedocs/gcc-14.1.0/gcc/Invoking-Gcov.html>
"""
# pylint: disable=too-many-lines

import os
from typing import Any, Iterator

from gcovr.utils import get_md5_hexdigest, read_source_file

from ....data_model.coverage import FileCoverage
from ....data_model.merging import FUNCTION_MAX_LINE_MERGE_OPTIONS, MergeOptions
from ....filter import Filter, is_file_excluded
from ....logging import LOGGER
from .common import (
    SUSPICIOUS_COUNTER,
    check_hits,
)

GCOV_JSON_VERSION = "2"


def parse_coverage(
    data_fname: str,
    gcov_json_data: dict[str, Any],
    *,
    include_filter: tuple[Filter, ...],
    exclude_filter: tuple[Filter, ...],
    source_encoding: str,
    ignore_parse_errors: set[str] | None,
    suspicious_hits_threshold: int = SUSPICIOUS_COUNTER,
    activate_trace_logging: bool = False,
) -> Iterator[tuple[FileCoverage, list[str]]]:
    """Process a GCOV JSON output."""

    # Check format version because the file can be created external
    if gcov_json_data["format_version"] != GCOV_JSON_VERSION:
        raise RuntimeError(
            f"Got wrong JSON format version {gcov_json_data['format_version']}, expected {GCOV_JSON_VERSION}"
        )

    for file in gcov_json_data["files"]:
        if not file["lines"] and not file["functions"]:
            if activate_trace_logging:
                LOGGER.trace(
                    "Skip data for file %s because no lines and functions defined.",
                    file["file"],
                )
            continue

        fname = os.path.normpath(
            os.path.join(gcov_json_data["current_working_directory"], file["file"])
        )

        if is_file_excluded("source file", fname, include_filter, exclude_filter):
            continue

        LOGGER.debug("Parsing coverage data for file %s", fname)

        max_line_number = (
            max(line["line_number"] for line in file["lines"]) if file["lines"] else 1
        )
        encoded_source_lines = read_source_file(source_encoding, fname, max_line_number)

        yield (
            _parse_file_node(
                data_fname,
                gcov_file_node=file,
                filename=fname,
                source_lines=encoded_source_lines,
                ignore_parse_errors=ignore_parse_errors,
                suspicious_hits_threshold=suspicious_hits_threshold,
                activate_trace_logging=activate_trace_logging,
            ),
            encoded_source_lines,
        )


def _parse_file_node(
    data_fname: str,
    *,
    gcov_file_node: dict[str, Any],
    filename: str,
    source_lines: list[str],
    ignore_parse_errors: set[str] | None,
    suspicious_hits_threshold: int = SUSPICIOUS_COUNTER,
    activate_trace_logging: bool = False,
) -> FileCoverage:
    """
    Extract coverage data from a json gcov report.

    Logging:
    Parse problems are reported as warnings.
    Coverage exclusion decisions are reported as verbose messages.

    Arguments:
        gcov_file_node: one of the "files" node in the gcov json format
        filename: for error reports
        source_lines: decoded source code lines, for reporting
        data_fname: source of this node, for reporting
        ignore_parse_errors: which errors should be converted to warnings

    Returns:
        The coverage data

    Raises:
        Any exceptions during parsing, unless ignore_parse_errors is set.
    """
    persistent_states: dict[str, Any] = {"location": (filename, 0)}

    if ignore_parse_errors is None:
        ignore_parse_errors = set()

    filecov = FileCoverage(data_fname, filename=filename)
    for line in gcov_file_node["lines"]:
        persistent_states.update(location=(filename, line["line_number"]))
        if activate_trace_logging:
            LOGGER.trace(
                "Reading %s of function %s",
                ":".join(str(e) for e in persistent_states["location"]),
                line.get("function_name"),
            )
        linecov = filecov.insert_line_coverage(
            str(data_fname),
            lineno=line["line_number"],
            count=check_hits(
                line["count"],
                source_lines[line["line_number"] - 1],
                ignore_parse_errors,
                suspicious_hits_threshold,
                persistent_states,
            ),
            function_name=line.get("function_name"),
            block_ids=line["block_ids"],
            md5=get_md5_hexdigest(
                source_lines[line["line_number"] - 1].encode("UTF-8")
            ),
        )
        for branch in line["branches"]:
            linecov.insert_branch_coverage(
                str(data_fname),
                branchno=None,
                count=check_hits(
                    branch["count"],
                    source_lines[line["line_number"] - 1],
                    ignore_parse_errors,
                    suspicious_hits_threshold,
                    persistent_states,
                ),
                source_block_id=branch["source_block_id"],
                fallthrough=branch["fallthrough"],
                throw=branch["throw"],
                destination_block_id=branch["destination_block_id"],
            )
        for index, condition in enumerate(line.get("conditions", [])):
            linecov.insert_condition_coverage(
                str(data_fname),
                conditionno=index,
                count=check_hits(
                    condition["count"],
                    source_lines[line["line_number"] - 1],
                    ignore_parse_errors,
                    suspicious_hits_threshold,
                    persistent_states,
                ),
                covered=condition["covered"],
                not_covered_true=condition["not_covered_true"],
                not_covered_false=condition["not_covered_false"],
            )
        for call in line.get("calls", []):
            linecov.insert_call_coverage(
                str(data_fname),
                callno=None,
                source_block_id=call["source_block_id"],
                destination_block_id=call["destination_block_id"],
                returned=call["returned"],
            )

    for function in gcov_file_node["functions"]:
        # Use 100% only if covered == total.
        if function["blocks_executed"] == function["blocks"]:
            blocks = 100.0
        else:
            # There is at least one uncovered item.
            # Round to 1 decimal and clamp to max 99.9%.
            ratio = function["blocks_executed"] / function["blocks"]
            blocks = min(99.9, round(ratio * 100.0, 1))

        filecov.insert_function_coverage(
            str(data_fname),
            MergeOptions(func_opts=FUNCTION_MAX_LINE_MERGE_OPTIONS),
            mangled_name=function["name"],
            demangled_name=function["demangled_name"],
            lineno=function["start_line"],
            count=function["execution_count"],
            blocks=blocks,
            start=(function["start_line"], function["start_column"]),
            end=(function["end_line"], function["end_column"]),
        )

    if (
        "negative_hits.warn_once_per_file" in persistent_states
        and persistent_states["negative_hits.warn_once_per_file"] > 1
    ):
        LOGGER.warning(
            "Ignored %d negative hits overall.",
            persistent_states["negative_hits.warn_once_per_file"],
        )

    if (
        "suspicious_hits.warn_once_per_file" in persistent_states
        and persistent_states["suspicious_hits.warn_once_per_file"] > 1
    ):
        LOGGER.warning(
            "Ignored %d suspicious hits overall.",
            persistent_states["suspicious_hits.warn_once_per_file"],
        )

    return filecov



# -*- coding:utf-8 -*-

#  ************************** Copyrights and license ***************************
#
# This file is part of gcovr 8.6, a parsing and reporting tool for gcov.
# https://gcovr.com/en/8.6
#
# _____________________________________________________________________________
#
# Copyright (c) 2013-2026 the gcovr authors
# Copyright (c) 2013 Sandia Corporation.
# Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
# the U.S. Government retains certain rights in this software.
#
# This software is distributed under the 3-clause BSD License.
# For more information, see the README.rst file.
#
# ****************************************************************************

from dataclasses import dataclass, field

from ..options import Options


@dataclass
class MergeFunctionOptions:
    """Data class to store the function merge options."""

    ignore_function_lineno: bool = False
    merge_function_use_line_zero: bool = False
    merge_function_use_line_min: bool = False
    merge_function_use_line_max: bool = False
    separate_function: bool = False


FUNCTION_STRICT_MERGE_OPTIONS = MergeFunctionOptions()
FUNCTION_LINE_ZERO_MERGE_OPTIONS = MergeFunctionOptions(
    ignore_function_lineno=True,
    merge_function_use_line_zero=True,
)
FUNCTION_MIN_LINE_MERGE_OPTIONS = MergeFunctionOptions(
    ignore_function_lineno=True,
    merge_function_use_line_min=True,
)
FUNCTION_MAX_LINE_MERGE_OPTIONS = MergeFunctionOptions(
    ignore_function_lineno=True,
    merge_function_use_line_max=True,
)
SEPARATE_FUNCTION_MERGE_OPTIONS = MergeFunctionOptions(
    ignore_function_lineno=True,
    separate_function=True,
)


@dataclass
class MergeOptions:
    """Data class to store the merge options."""

    func_opts: MergeFunctionOptions = field(default_factory=MergeFunctionOptions)


DEFAULT_MERGE_OPTIONS = MergeOptions()


def get_merge_mode_from_options(options: Options) -> MergeOptions:
    """Get the function merge mode."""
    merge_opts = MergeOptions()
    if options.merge_mode_functions == "strict":
        merge_opts.func_opts = FUNCTION_STRICT_MERGE_OPTIONS
    elif options.merge_mode_functions == "merge-use-line-0":
        merge_opts.func_opts = FUNCTION_LINE_ZERO_MERGE_OPTIONS
    elif options.merge_mode_functions == "merge-use-line-min":
        merge_opts.func_opts = FUNCTION_MIN_LINE_MERGE_OPTIONS
    elif options.merge_mode_functions == "merge-use-line-max":
        merge_opts.func_opts = FUNCTION_MAX_LINE_MERGE_OPTIONS
    elif options.merge_mode_functions == "separate":
        merge_opts.func_opts = SEPARATE_FUNCTION_MERGE_OPTIONS
    else:
        raise AssertionError("Unknown functions merge mode.")

    return merge_opts



# -*- coding:utf-8 -*-

#  ************************** Copyrights and license ***************************
#
# This file is part of gcovr 8.6, a parsing and reporting tool for gcov.
# https://gcovr.com/en/8.6
#
# _____________________________________________________________________________
#
# Copyright (c) 2013-2026 the gcovr authors
# Copyright (c) 2013 Sandia Corporation.
# Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
# the U.S. Government retains certain rights in this software.
#
# This software is distributed under the 3-clause BSD License.
# For more information, see the README.rst file.
#
# ****************************************************************************

import logging
import os
from ...data_model.container import CoverageContainer
from ...formats.base import BaseHandler
from ...options import GcovrConfigOption

LOGGER = logging.getLogger("gcovr")


class LlvmHandler(BaseHandler):
    """Class to handle LLVM JSON tracefile format."""

    @classmethod
    def get_options(cls) -> list[GcovrConfigOption | str]:
        return [
            # Global options needed for processing
            "keep_intermediate_files",
            "delete_input_files",
            # Global options needed for report
            "show_decision",
            # Global options used for merging end exclusion processing.
            "exclude_directory",
            "exclude_noncode_lines",
            "exclude_throw_branches",
            "exclude_unreachable_branches",
            "exclude_function_lines",
            "exclude_internal_functions",
            "respect_exclusion_markers",
            "exclude_function",
            "exclude_lines_by_pattern",
            "exclude_branches_by_pattern",
            "exclude_pattern_prefix",
            "warn_excluded_lines_with_hits",
            "merge_mode_functions",
            # Local options
            GcovrConfigOption(
                "llvm_profdata_cmd",
                ["--llvm-profdata-executable"],
                group="llvm_options",
                help=(
                    "Use a particular llvm-profdata executable to convert LLVM profraw files. "
                    "This switches from searching gcno/gcda files and using gcov to searching "
                    "profraw files (Source-based Code Coverage) of LLVM. "
                    "Must match the compiler you are using, e.g. llvm-profdata-13 for clang-13. "
                    "Defaults to the LLVM_PROFDATA environment variable: '{default!s}'."
                ),
                default=os.environ.get("LLVM_PROFDATA"),
            ),
            GcovrConfigOption(
                "llvm_cov_binaries",
                ["--llvm-cov-binary"],
                group="llvm_options",
                help=(
                    "The binary to export the coverage data for. See help of 'llvm-cov export' command. "
                    "The option can be used multiple times."
                ),
                type=str,
                action="append",
            ),
        ]

    def read_report(self) -> CoverageContainer:
        from .read import read_report  # pylint: disable=import-outside-toplevel # Lazy loading is intended here

        # This can't be checked during validation of arguments because it's not needed together with the
        # tracefiles
        if not self.options.llvm_cov_binaries:
            raise RuntimeError(
                "Missing --llvm-cov-binary, needed for llvm-cov execution."
            )

        return read_report(self.options)



# -*- coding:utf-8 -*-

#  ************************** Copyrights and license ***************************
#
# This file is part of gcovr 8.6, a parsing and reporting tool for gcov.
# https://gcovr.com/en/8.6
#
# _____________________________________________________________________________
#
# Copyright (c) 2013-2026 the gcovr authors
# Copyright (c) 2013 Sandia Corporation.
# Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
# the U.S. Government retains certain rights in this software.
#
# This software is distributed under the 3-clause BSD License.
# For more information, see the README.rst file.
#
# ****************************************************************************

import gzip
from hashlib import md5
import json
import lzma
from typing import Any, Callable, Iterator
import os
import functools
import re
import sys
from contextlib import contextmanager
from lxml import etree  # nosec # We only write XML files

from .logging import LOGGER
from .version import __version__

REGEX_VERSION_POSTFIX = re.compile(r"(.+?)(?:\.post\d+)?\.dev.+$")
PRETTY_JSON_INDENT = 4
GZIP_SUFFIX = ".gz"
LZMA_SUFFIX = ".xz"


class LoopChecker:
    """Class for checking if a directory was already scanned."""

    def __init__(self) -> None:
        self._seen = set[tuple[int, int]]()

    def already_visited(self, path: str) -> bool:
        """Check if the path was already checked."""
        st = os.stat(path)
        key = (st.st_dev, st.st_ino)
        if key in self._seen:
            return True

        self._seen.add(key)
        return False


@functools.lru_cache(maxsize=1)
def is_fs_case_insensitive() -> bool:
    """Check if the file system is case insensitive."""
    cwd = os.getcwd()
    # Guessing if file system is case insensitive.
    # The working directory is not the root and accessible in upper and lower case
    # and pointing to same file.
    ret = (
        (cwd != os.path.sep)
        and os.path.exists(cwd.upper())
        and os.path.exists(cwd.lower())
        and os.path.samefile(cwd.upper(), cwd.lower())
    )
    LOGGER.debug("File system is case %s.", "insensitive" if ret else "sensitive")

    return ret


@functools.lru_cache(maxsize=None)
def fix_case_of_path(path: str) -> str:
    """Fix casing of filenames for cas insensitive file systems."""
    rest, cur = os.path.split(path)
    # e.g path = ".." happens if original path is like "../dir/subdir/file.cpp"
    if not rest:
        return cur
    if not cur:  # e.g path = "C:/"
        return rest.upper()  # Always use uppercase drive letter

    try:
        cur_lower = cur.lower()
        matched_filename = [f for f in os.listdir(rest) if f.lower() == cur_lower]
        if len(matched_filename) > 1:
            raise RuntimeError(
                "Seems that we have a case sensitive filesystem, can't fix file case"
            )

        if len(matched_filename) == 1:
            path = os.path.join(fix_case_of_path(rest), matched_filename[0])
    except FileNotFoundError:
        LOGGER.warning("Can not fix case of path because %s not found.", rest)

    return path.replace("\\", "/")


def get_version_for_report() -> str:
    """Get the printable version for the report."""
    version = __version__
    if match := REGEX_VERSION_POSTFIX.match(version):
        version = f"{match.group(1)}+main"
    return version


def search_file(
    predicate: Callable[[str], bool],
    path: str,
    exclude_directory: list[re.Pattern[str]],
) -> Iterator[str]:
    """
    Given a search path, recursively descend to find files that satisfy a
    predicate.
    """
    if path is None or path == ".":
        path = os.getcwd()
    elif not os.path.exists(path):
        raise IOError("Unknown directory '" + path + "'")

    loop_checker = LoopChecker()
    for root, dirs, files in os.walk(os.path.abspath(path), followlinks=True):
        # Check if we've already visited 'root' through the magic of symlinks
        if loop_checker.already_visited(root):
            dirs[:] = []
            continue

        dirs[:] = [
            d
            for d in sorted(dirs)
            if not any(exc.match(os.path.join(root, d)) for exc in exclude_directory)
        ]
        root = os.path.abspath(root)

        for name in sorted(files):
            if predicate(name):
                yield os.path.abspath(os.path.join(root, name))


def commonpath(files: list[str]) -> str:
    r"""Find the common prefix of all files.

    This differs from the standard library os.path.commonpath():
     - We first normalize all paths to a realpath.
     - We return a path with a trailing path separator.

    No common path exists under the following circumstances:
     - on Windows when the paths have different drives.
       E.g.: commonpath([r'C:\foo', r'D:\foo']) == ''
     - when the `files` are empty.

    Arguments:
        files (list): the input paths, may be relative or absolute.

    Returns: str
        The common prefix directory as a relative path.
        Always ends with a path separator.
        Returns the empty string if no common path exists.
    """
    if not files:
        return ""

    if len(files) == 1:
        prefix_path = str(os.path.dirname(os.path.realpath(files[0])))
    else:
        split_paths = [os.path.realpath(path).split(os.path.sep) for path in files]
        # We only have to compare the lexicographically minimum and maximum
        # paths to find the common prefix of all, e.g.:
        #   /a/b/c/d  <- min
        #   /a/b/d
        #   /a/c/a    <- max
        #
        # compare:
        # https://github.com/python/cpython/blob/3.6/Lib/posixpath.py#L487
        min_path = min(split_paths)
        max_path = max(split_paths)
        common = min_path  # assume that min_path is a prefix of max_path
        for i in range(min(len(min_path), len(max_path))):
            if min_path[i] != max_path[i]:
                common = min_path[:i]  # disproven, slice for actual prefix
                break
        prefix_path = os.path.sep.join(common)

    LOGGER.debug("Common prefix path is %r.", prefix_path)

    # make the path relative and add a trailing slash
    if prefix_path:
        prefix_path = os.path.join(
            os.path.relpath(prefix_path, os.path.realpath(os.getcwd())), ""
        )
        LOGGER.debug("Common relative prefix path is %r.", prefix_path)
    return prefix_path


@contextmanager
def open_text_for_writing(
    filename: str | None, default_filename: str | None = None, **kwargs: Any
) -> Iterator[Any]:
    """Context manager to open and close a file for text writing.

    Stdout is used if `filename` is None or '-'.
    """
    if filename is not None and filename.endswith(os.sep):
        if default_filename is None:
            raise AssertionError(
                "If filename is a directory a default filename is mandatory."
            )
        filename += default_filename

    if filename is not None and filename != "-":
        if filename.endswith(GZIP_SUFFIX):
            with gzip.open(filename, "wt", **kwargs) as fh_out:
                yield fh_out
        elif filename.endswith(LZMA_SUFFIX):
            with lzma.open(filename, "wt", **kwargs) as fh_out:
                yield fh_out
        else:
            with open(filename, "wt", **kwargs) as fh_out:  # pylint: disable=unspecified-encoding
                yield fh_out
    else:
        encoding = kwargs.get("encoding", "utf-8").lower()
        old_encoding = sys.stdout.encoding
        try:
            if old_encoding != encoding:
                sys.stdout.reconfigure(encoding=encoding)  # type: ignore[union-attr]
            yield sys.stdout
        finally:
            if old_encoding != encoding:
                sys.stdout.reconfigure(encoding=old_encoding)  # type: ignore[union-attr]


@contextmanager
def open_binary_for_writing(
    filename: str | None,
    default_filename: str,
    **kwargs: Any,
) -> Iterator[Any]:
    """Context manager to open and close a file for binary writing.

    Stdout is used if `filename` is None or '-'.
    """
    if filename is not None and filename.endswith(os.sep):
        filename += default_filename

    if filename is not None and filename != "-":
        if filename.endswith(GZIP_SUFFIX):
            with gzip.open(filename, "wb", **kwargs) as fh_out:
                yield fh_out
        elif filename.endswith(LZMA_SUFFIX):
            with lzma.open(filename, "wb", **kwargs) as fh_out:
                yield fh_out
        else:
            # files in write binary mode for UTF-8
            with open(filename, "wb", **kwargs) as fh_out:
                yield fh_out
    else:
        yield sys.stdout.buffer


def write_json_output(
    json_dict: dict[str, Any],
    *,
    pretty: bool,
    filename: str | None,
    default_filename: str,
    **kwargs: Any,
) -> None:
    """Helper function to output JSON dictionary to a file/STDOUT."""
    with open_text_for_writing(filename, default_filename, **kwargs) as fh:
        json.dump(json_dict, fh, indent=PRETTY_JSON_INDENT if pretty else None)


def write_xml_output(
    root: Any,
    *,
    doctype: str | None = None,
    pretty: bool,
    filename: str | None,
    default_filename: str,
    **kwargs: Any,
) -> None:
    """Helper function to output XML format dictionary to a file/STDOUT"""
    with open_binary_for_writing(filename, default_filename, **kwargs) as fh:
        fh.write(
            etree.tostring(
                root,
                pretty_print=pretty,
                encoding="utf-8",
                xml_declaration=True,
                doctype=doctype,  # type: ignore [arg-type]
            )
        )


@contextmanager
def chdir(directory: str) -> Iterator[None]:
    """Context for doing something in a locked directory."""
    current_dir = os.getcwd()
    os.chdir(directory)
    try:
        yield
    finally:
        os.chdir(current_dir)


def force_unix_separator(path: str) -> str:
    """Get the filename with / independent from OS."""
    return path.replace("\\", "/")


def get_md5_hexdigest(data: bytes) -> str:
    """Get the MD5 digest of the given bytes."""
    return md5(data, usedforsecurity=False).hexdigest()  # nosec # Not used for security


def read_source_file(
    source_encoding: str, filename: str, max_line_number: int
) -> list[str]:
    """Read in the source file and fill up lines if needed."""
    source_lines: list[bytes] = []
    try:
        with open(filename, "rb") as fh_in:
            source_lines = fh_in.read().splitlines()
        lines = len(source_lines)
        if lines < max_line_number:
            LOGGER.warning(
                "File %s has %d line(s) but coverage data has %d line(s).",
                filename,
                lines,
                max_line_number,
            )
            # GCOV itself adds the /*EOF*/ in the text report if there is no data and we used the same.
            source_lines += [b"/*EOF*/"] * (max_line_number - lines)
    except OSError as e:
        if filename.endswith("<stdin>"):
            message = (
                f"Got unreadable source file '{filename}', replacing with empty lines."
            )
            LOGGER.info(message)
        else:
            # The exception contains the source file name,
            # e.g. [Errno 2] No such file or directory: 'xy.txt'
            message = f"Can't read file, using empty lines: {e}"
            LOGGER.warning(message)
            # If we can't read the file we use as first line the error
            # and use empty lines for the rest of the lines.
        source_lines = [b""] * max_line_number
        source_lines[0] = f"/* {message} */".encode()

    encoded_source_lines = [
        line.decode(source_encoding, errors="replace") for line in source_lines
    ]

    return encoded_source_lines



# -*- coding:utf-8 -*-

#  ************************** Copyrights and license ***************************
#
# This file is part of gcovr 8.6, a parsing and reporting tool for gcov.
# https://gcovr.com/en/8.6
#
# _____________________________________________________________________________
#
# Copyright (c) 2013-2026 the gcovr authors
# Copyright (c) 2013 Sandia Corporation.
# Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
# the U.S. Government retains certain rights in this software.
#
# This software is distributed under the 3-clause BSD License.
# For more information, see the README.rst file.
#
# ****************************************************************************

from __future__ import annotations
from typing import TypeVar
from dataclasses import dataclass

from ..options import Options


_T = TypeVar("_T")


@dataclass
class SummarizedStats:
    """Data class for the summarized coverage statistics."""

    line: CoverageStat
    branch: CoverageStat
    condition: CoverageStat
    decision: DecisionCoverageStat
    function: CoverageStat
    call: CoverageStat

    @staticmethod
    def new_empty() -> SummarizedStats:
        """Create a empty coverage statistic."""
        return SummarizedStats(
            line=CoverageStat.new_empty(),
            branch=CoverageStat.new_empty(),
            condition=CoverageStat.new_empty(),
            decision=DecisionCoverageStat.new_empty(),
            function=CoverageStat.new_empty(),
            call=CoverageStat.new_empty(),
        )

    def __iadd__(self, other: SummarizedStats) -> SummarizedStats:
        self.line += other.line
        self.branch += other.branch
        self.condition += other.condition
        self.decision += other.decision
        self.function += other.function
        self.call += other.call
        return self

    def serialize(
        self, default_percent: _T, options: Options
    ) -> dict[str, int | float | _T]:
        """Serialize the object."""
        data_dict = dict[str, int | float | _T](
            {
                "line_total": self.line.total,
                "line_covered": self.line.covered,
                "line_percent": self.line.percent_or(default_percent),
                "function_total": self.function.total,
                "function_covered": self.function.covered,
                "function_percent": self.function.percent_or(default_percent),
                "branch_total": self.branch.total,
                "branch_covered": self.branch.covered,
                "branch_percent": self.branch.percent_or(default_percent),
            }
        )
        if self.condition.total != 0:
            data_dict.update(
                {
                    "condition_total": self.condition.total,
                    "condition_covered": self.condition.covered,
                    "condition_percent": self.condition.percent_or(default_percent),
                }
            )

        if options.show_decision:
            data_dict.update(
                {
                    "decision_total": self.decision.total,
                    "decision_covered": self.decision.covered,
                    "decision_percent": self.decision.percent_or(default_percent),
                }
            )

        return data_dict


@dataclass
class CoverageStat:
    """A single coverage metric, e.g. the line coverage percentage of a file."""

    covered: int
    """How many elements were covered."""

    excluded: int
    """How many elements there were excluded."""

    total_with_excluded: int
    """How many elements there were in total (including the excluded ones)."""

    @staticmethod
    def new_empty() -> CoverageStat:
        """Create a empty coverage statistic."""
        return CoverageStat(0, 0, 0)

    @property
    def total(self) -> int:
        """Get the total without the excluded elements."""
        return self.total_with_excluded - self.excluded

    @property
    def percent(self) -> float | None:
        """Percentage of covered elements, equivalent to ``self.percent_or(None)``"""
        return self.percent_or(None)

    def percent_or(self, default: _T) -> float | _T:
        """Percentage of covered elements.

        Coverage is truncated to one decimal:
        >>> CoverageStat(1234, 0, 10000).percent_or("default")
        12.3

        Coverage is capped at 99.9% unless everything is covered:
        >>> CoverageStat(9999, 0, 10000).percent_or("default")
        99.9
        >>> CoverageStat(10000, 0, 10000).percent_or("default")
        100.0

        If there are no elements, percentage is NaN and the default will be returned:
        >>> CoverageStat(0, 0, 0).percent_or("default")
        'default'
        """
        if not self.total:
            return default

        # Return 100% only if covered == total.
        if self.covered == self.total:
            return 100.0

        # There is at least one uncovered item.
        # Round to 1 decimal and clamp to max 99.9%.
        ratio = self.covered / self.total
        return min(99.9, round(ratio * 100.0, 1))

    def __iadd__(self, other: CoverageStat) -> CoverageStat:
        self.covered += other.covered
        self.total_with_excluded += other.total_with_excluded
        self.excluded += other.excluded
        return self


@dataclass
class DecisionCoverageStat:
    """A CoverageStat for decision coverage (accounts for Uncheckable cases)."""

    covered: int
    uncheckable: int
    total: int

    @classmethod
    def new_empty(cls) -> DecisionCoverageStat:
        """Create a empty decision coverage statistic."""
        return cls(0, 0, 0)

    @property
    def to_coverage_stat(self) -> CoverageStat:
        """Convert a decision coverage statistic to a coverage statistic."""
        return CoverageStat(
            covered=self.covered, excluded=0, total_with_excluded=self.total
        )

    @property
    def percent(self) -> float | None:
        """Return the percent value of the coverage."""
        return self.to_coverage_stat.percent

    def percent_or(self, default: _T) -> float | _T:
        """Return the percent value of the coverage or the given default if no coverage is present."""
        return self.to_coverage_stat.percent_or(default)

    def __iadd__(self, other: DecisionCoverageStat) -> DecisionCoverageStat:
        self.covered += other.covered
        self.uncheckable += other.uncheckable
        self.total += other.total
        return self



# -*- coding:utf-8 -*-

#  ************************** Copyrights and license ***************************
#
# This file is part of gcovr 8.6, a parsing and reporting tool for gcov.
# https://gcovr.com/en/8.6
#
# _____________________________________________________________________________
#
# Copyright (c) 2013-2026 the gcovr authors
# Copyright (c) 2013 Sandia Corporation.
# Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
# the U.S. Government retains certain rights in this software.
#
# This software is distributed under the 3-clause BSD License.
# For more information, see the README.rst file.
#
# ****************************************************************************

import os
import re
import sys

from argparse import ArgumentError, ArgumentParser, Namespace
from typing import Any
import traceback

from .configuration import (
    argument_parser_setup,
    config_entries_from_dict,
    merge_options_and_set_defaults,
    parse_config_file,
    parse_config_into_dict,
)
from .data_model.container import CoverageContainer
from .exceptions import SanityCheckError
from .filter import (
    AlwaysMatchFilter,
    DirectoryPrefixFilter,
    Filter,
)
from .formats.gcov.read import GcovProgram
from .formats.gcov.workers import Workers
from .logging import configure_logging, update_logging, LOGGER
from .options import FilterOption
from .version import __version__

# formats
from . import formats as gcovr_formats

if sys.version_info >= (3, 11):
    import tomllib
else:
    import tomli as tomllib

EXIT_SUCCESS = 0
EXIT_CMDLINE_ERROR = 1
EXIT_LINE_NOK = 2
EXIT_BRANCH_NOK = 4
EXIT_DECISION_NOK = 8
EXIT_FUNCTION_NOK = 16
EXIT_READ_ERROR = 64
EXIT_WRITE_ERROR = 128


#
# Exits with status 2 if below threshold
#
def get_exit_code(
    covdata: CoverageContainer,
    threshold_line: float,
    threshold_branch: float,
    threshold_decision: float,
    threshold_function: float,
) -> int:
    """Fail depending on the coverage result."""
    exit_code = 0

    if (
        threshold_line > 0.0
        or threshold_branch > 0.0
        or threshold_decision > 0.0
        or threshold_function > 0.0
    ):
        stats = covdata.stats

        line_nok = False
        if threshold_line > 0.0:
            # If there are no lines, mark as uncovered
            # (indicates no data at all, likely an error).
            percent_lines = stats.line.percent_or(0.0)

            if percent_lines < threshold_line:
                line_nok = True
                LOGGER.error(
                    "Failed minimum line coverage (got %s%%, minimum %s%%)",
                    percent_lines,
                    threshold_line,
                )

        branch_nok = False
        if threshold_branch > 0.0:
            # Allow data with no branches.
            percent_branches = stats.branch.percent_or(100.0)
            if percent_branches < threshold_branch:
                branch_nok = True
                LOGGER.error(
                    "Failed minimum branch coverage (got %s%%, minimum %s%%)",
                    percent_branches,
                    threshold_branch,
                )

        decision_nok = False
        if threshold_decision > 0.0:
            # Allow data with no decisions.
            percent_decision = stats.decision.percent_or(100.0)
            if percent_decision < threshold_decision:
                decision_nok = True
                LOGGER.error(
                    "Failed minimum decision coverage (got %s%%, minimum %s%%)",
                    percent_decision,
                    threshold_decision,
                )

        function_nok = False
        if threshold_function > 0.0:
            # Allow data with no functions.
            percent_function = stats.function.percent_or(100.0)
            if percent_function < threshold_function:
                function_nok = True
                LOGGER.error(
                    "Failed minimum function coverage (got %s%%, minimum %s%%)",
                    percent_function,
                    threshold_function,
                )

        if line_nok:
            exit_code |= EXIT_LINE_NOK
        if branch_nok:
            exit_code |= EXIT_BRANCH_NOK
        if decision_nok:
            exit_code |= EXIT_DECISION_NOK
        if function_nok:
            exit_code |= EXIT_FUNCTION_NOK

    return exit_code


def create_argument_parser() -> ArgumentParser:
    """Create the argument parser."""

    parser = ArgumentParser(add_help=False, exit_on_error=False)
    parser.usage = "gcovr [options] [search_paths...]"
    parser.description = (
        "A utility to run gcov and summarize the coverage in simple reports."
    )

    parser.epilog = "See <http://gcovr.com/> for the full manual."

    options = parser.add_argument_group("Options")
    options.add_argument(
        "-h", "--help", help="Show this help message, then exit.", action="help"
    )
    options.add_argument(
        "--version",
        help="Print the version number, then exit.",
        action="store_true",
        dest="version",
        default=False,
    )

    argument_parser_setup(parser, options)

    return parser


COPYRIGHT = (
    "Copyright (c) 2013-2026 the gcovr authors\n"
    "Copyright (c) 2013 Sandia Corporation.\n"
    "Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,\n"
    "the U.S. Government retains certain rights in this software.\n"
)


def find_config_name(root: str, filename: str) -> str | None:
    """Find the configuration to use."""
    if root:
        filename = os.path.join(root, filename)

    if os.path.isfile(filename):
        return filename

    return None


def load_config(partial_options: Namespace) -> dict[str, Any]:
    """Load a config file if configured or found by default names"""
    filename = getattr(partial_options, "config", None)
    if filename is not None:
        with open(filename, encoding="utf-8") as buf:
            return parse_config_into_dict(parse_config_file(buf, filename))

    root = getattr(partial_options, "root", "")
    if filename := find_config_name(root, "gcovr.cfg"):
        with open(filename, encoding="utf-8") as buf:
            return parse_config_into_dict(parse_config_file(buf, filename))

    if filename := find_config_name(root, "gcovr.toml"):
        with open(filename, "rb") as buf:
            data = tomllib.load(buf)
        return parse_config_into_dict(config_entries_from_dict(data, filename))

    if filename := find_config_name(root, "pyproject.toml"):
        with open(filename, "rb") as buf:
            data = tomllib.load(buf)
        if (gcovr_section := data.get("tool", {}).get("gcovr")) is not None:
            return parse_config_into_dict(
                config_entries_from_dict(gcovr_section, filename)
            )

    return {}


def main(args: list[str] | None = None) -> int:  # pylint: disable=too-many-return-statements
    """The main entry point of GCOVR."""
    configure_logging()
    try:
        parser = create_argument_parser()
        cli_options = parser.parse_args(args=args)
    except SystemExit as e:
        if e.code != 0:
            raise SanityCheckError("Exitcode must be 0.") from e
        return EXIT_SUCCESS
    except ArgumentError as e:
        sys.stderr.write(f"gcovr: error: {e}\n")
        return EXIT_CMDLINE_ERROR

    if cli_options.version:
        sys.stdout.write(f"gcovr {__version__}\n\n{COPYRIGHT}")
        return EXIT_SUCCESS

    # load the config
    cfg_options = load_config(cli_options)
    options = merge_options_and_set_defaults([cfg_options, cli_options.__dict__])

    # Reconfigure the logging.
    update_logging(options)
    LOGGER.debug("gcovr version %s", __version__)

    # We need to reset the stored information her for our test framework
    GcovProgram.reset()

    if options.sort_branches and options.sort_key not in [
        "uncovered-number",
        "uncovered-percent",
    ]:
        LOGGER.error(
            "the options --sort-branches without '--sort uncovered-number' or '--sort uncovered-percent' doesn't make sense."
        )
        return EXIT_CMDLINE_ERROR

    if options.html_title == "":
        LOGGER.error("an empty --html-title= is not allowed.")
        return EXIT_CMDLINE_ERROR

    for postfix in ["", "line", "branch"]:
        key_medium = "medium_threshold"
        key_high = "high_threshold"
        if postfix:
            key_medium += f"_{postfix}"
            key_high += f"_{postfix}"
        option_medium = f"--{key_medium.replace('_', '-')}"
        option_high = f"--{key_high.replace('_', '-')}"

        if getattr(options, key_medium) == 0:
            LOGGER.error("Value of %s should not be zero.", option_medium)
            return EXIT_CMDLINE_ERROR

        # Inherit the defaults from the global coverage values if not set
        if postfix:
            if getattr(options, key_medium) is None:
                setattr(
                    options,
                    key_medium,
                    options.medium_threshold,
                )
                # To get the correct option in the error message below.
                option_medium = "--medium-threshold"
            if getattr(options, key_high) is None:
                setattr(
                    options,
                    key_high,
                    options.high_threshold,
                )
                # To get the correct option in the error message below.
                option_medium = "--high-threshold"

        if getattr(options, key_medium) > getattr(options, key_high):
            LOGGER.error(
                "Value of %s=%s should be\nlower than or equal to the value of %s=%s.",
                option_medium,
                getattr(options, key_medium),
                option_high,
                getattr(options, key_high),
            )
            return EXIT_CMDLINE_ERROR

    try:
        gcovr_formats.validate_options(options)
    except RuntimeError as exc:
        LOGGER.error("%s", str(exc))
        return EXIT_CMDLINE_ERROR

    options.starting_dir = os.path.abspath(os.getcwd())
    options.root_dir = os.path.abspath(options.root)

    #
    # Setup filters
    #
    # The root filter isn't technically a filter,
    # but is used to turn absolute paths into relative paths
    options.root_filter = re.compile("^" + re.escape(options.root_dir + os.sep))

    def _setup_filter(
        option: str,
        patterns: list[FilterOption],
        default_filter: Filter | None = None,
    ) -> tuple[Filter, ...]:
        """Setup a filter and handle the exception."""
        try:
            filters = list[Filter]()
            if len(patterns):
                filters = [f.build_filter() for f in patterns]
            elif default_filter is not None:
                filters.append(default_filter)
            LOGGER.debug("Filters for %s: (%d)", option, len(filters))
            for f in filters:
                LOGGER.debug(" - %s", f)
            return tuple(filters)
        except re.error as e:
            # mypy is thinking that the pattern can be a byte string therefore we need to explicit use !s.
            # See also discussion https://github.com/gcovr/gcovr/pull/1028#discussion_r1855437452
            raise RuntimeError(
                f"Error setting up filter {option}='{e.pattern!s}': {e}"
            ) from None

    def _setup_pattern(option: str, patterns: list[str]) -> tuple[re.Pattern[str], ...]:
        """Setup a filter and handle the exception."""
        try:
            compiled_patterns = list[re.Pattern[str]]()
            if len(patterns):
                compiled_patterns = [re.compile(p) for p in patterns]
            LOGGER.debug("Patterns for %s: (%d)", option, len(compiled_patterns))
            for p in compiled_patterns:
                LOGGER.debug(" - %s", p)
            return tuple(compiled_patterns)
        except re.error as e:
            # mypy is thinking that the pattern can be a byte string therefore we need to explicit use !s.
            # See also discussion https://github.com/gcovr/gcovr/pull/1028#discussion_r1855437452
            raise RuntimeError(
                f"Error setting up pattern {option}='{e.pattern!s}': {e}"
            ) from None

    try:
        options.include_filter = _setup_filter(
            "--filter", options.include_filter, DirectoryPrefixFilter(options.root_dir)
        )
        options.exclude_filter = _setup_filter("--exclude", options.exclude_filter)
        options.include_search_filter = _setup_filter(
            "--include", options.include_search_filter
        )

        options.gcov_include_filter = _setup_filter(
            "--gcov-filter", options.gcov_include_filter, AlwaysMatchFilter()
        )
        options.gcov_exclude_filter = _setup_filter(
            "--gcov-exclude", options.gcov_exclude_filter
        )
        options.exclude_directory = _setup_filter(
            "--exclude-directory", options.exclude_directory
        )

        options.trace_include_filter = _setup_filter(
            "--trace-include", options.trace_include_filter
        )
        options.trace_exclude_filter = _setup_filter(
            "--trace-exclude", options.trace_exclude_filter
        )

        options.exclude_function = _setup_pattern(
            "--exclude-function",
            [
                p[1:-1] if p[0] == "/" and p[-1] == "/" else re.escape(p)
                for p in options.exclude_function
            ],
        )
        options.exclude_lines_by_pattern = _setup_pattern(
            "--exclude-lines-by-pattern", options.exclude_lines_by_pattern
        )
        options.exclude_branches_by_pattern = _setup_pattern(
            "--exclude-branches-by-pattern", options.exclude_branches_by_pattern
        )
    except RuntimeError as e:
        LOGGER.error("%s", e)
        return EXIT_CMDLINE_ERROR

    if options.fail_under_decision > 0.0 and not options.show_decision:
        LOGGER.error("--fail-under-decision need also option --decision.")
        return EXIT_CMDLINE_ERROR

    if options.show_decision:
        LOGGER.info(
            "Attention, the decision analysis is experimental. "
            "It uses a fragile heuristic and depends on the code format."
        )

    LOGGER.info("Reading coverage data...")
    try:
        covdata = gcovr_formats.read_reports(options)
    except Workers.WorkerThreadException as exc:
        LOGGER.error("Error occurred while reading reports: %s", exc)
        return EXIT_READ_ERROR
    except Exception:  # pylint: disable=broad-exception-caught
        LOGGER.error(
            "Error occurred while reading reports:\n%s", traceback.format_exc()
        )
        return EXIT_READ_ERROR

    LOGGER.info("Writing coverage report...")
    try:
        gcovr_formats.write_reports(covdata, options)
    except Exception:  # pylint: disable=broad-exception-caught
        LOGGER.error(
            "Error occurred while printing reports:\n%s", traceback.format_exc()
        )
        return EXIT_WRITE_ERROR

    return get_exit_code(
        covdata,
        options.fail_under_line,
        options.fail_under_branch,
        options.fail_under_decision,
        options.fail_under_function,
    )


if __name__ == "__main__":
    sys.exit(main())



# -*- coding:utf-8 -*-

#  ************************** Copyrights and license ***************************
#
# This file is part of gcovr 8.6, a parsing and reporting tool for gcov.
# https://gcovr.com/en/8.6
#
# _____________________________________________________________________________
#
# Copyright (c) 2013-2026 the gcovr authors
# Copyright (c) 2013 Sandia Corporation.
# Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
# the U.S. Government retains certain rights in this software.
#
# This software is distributed under the 3-clause BSD License.
# For more information, see the README.rst file.
#
# ****************************************************************************

from __future__ import annotations
import os
import re
from typing import Any, ItemsView, Iterable, Iterator, Literal, ValuesView

from ..filter import is_file_excluded

from ..logging import LOGGER
from ..options import Options
from ..utils import commonpath, force_unix_separator

from .coverage import FileCoverage
from .coverage_dict import CoverageDict
from .merging import MergeOptions
from .stats import CoverageStat, DecisionCoverageStat, SummarizedStats


class ContainerBase:
    """Base class for coverage containers"""

    def sort_coverage(
        self,
        sort_key: Literal["filename", "uncovered-number", "uncovered-percent"],
        sort_reverse: bool,
        by_metric: Literal["line", "branch", "decision"],
        filename_uses_relative_pathname: bool = False,
    ) -> list[str]:
        """Sort a coverage dict.

        covdata (dict): the coverage dictionary
        sort_key ("filename", "uncovered-number", "uncovered-percent"): the values to sort by
        sort_reverse (bool): reverse order if True
        by_metric ("line", "branch", "decision"): select the metric to sort
        filename_uses_relative_pathname (bool): for html, we break down a pathname to the
            relative path, but not for other formats.

        returns: the sorted keys
        """

        basedir = commonpath(list(self.data.keys()))

        def key_filename(key: str) -> list[int | str]:
            def convert_to_int_if_possible(text: str) -> int | str:
                return int(text) if text.isdigit() else text

            key = (
                force_unix_separator(
                    os.path.relpath(os.path.realpath(key), os.path.realpath(basedir))
                )
                if filename_uses_relative_pathname
                else key
            ).casefold()

            return [
                convert_to_int_if_possible(part) for part in re.split(r"([0-9]+)", key)
            ]

        def coverage_stat(key: str) -> CoverageStat:
            cov: FileCoverage | CoverageContainerDirectory = self.data[key]
            if by_metric == "branch":
                return cov.branch_coverage()
            if by_metric == "decision":
                return cov.decision_coverage().to_coverage_stat
            return cov.line_coverage()

        def key_num_uncovered(key: str) -> int:
            stat = coverage_stat(key)
            uncovered = stat.total - stat.covered
            return uncovered

        def key_percent_uncovered(key: str) -> float:
            stat = coverage_stat(key)
            covered = stat.covered
            total = stat.total

            # No branches are always put directly after (or before when reversed)
            # files with 100% coverage (by assigning such files 110% coverage)
            return covered / total if total > 0 else 1.1

        if sort_key == "uncovered-number":
            # First sort filename alphabetical and then by the requested key
            return sorted(
                sorted(self.data, key=key_filename),
                key=key_num_uncovered,
                reverse=sort_reverse,
            )
        if sort_key == "uncovered-percent":
            # First sort filename alphabetical and then by the requested key
            return sorted(
                sorted(self.data, key=key_filename),
                key=key_percent_uncovered,
                reverse=sort_reverse,
            )

        # By default, we sort by filename alphabetically
        return sorted(self.data, key=key_filename, reverse=sort_reverse)


class CoverageContainer(ContainerBase):
    """Coverage container holding all the coverage data."""

    def __init__(self) -> None:
        self.data = CoverageDict[str, FileCoverage]()
        self.directories = list[CoverageContainerDirectory]()

    def __getitem__(self, key: str) -> FileCoverage:
        return self.data[key]

    def __len__(self) -> int:
        return len(self.data)

    def __contains__(self, key: str) -> bool:
        return key in self.data

    def __iter__(self) -> Iterator[str]:
        return iter(self.data)

    def values(self) -> ValuesView[FileCoverage]:
        """Get the file coverage data objects."""
        return self.data.values()

    def items(self) -> ItemsView[str, FileCoverage]:
        """Get the file coverage data items."""
        return self.data.items()

    def serialize(self, options: Options) -> list[dict[str, Any]]:
        """Serialize the object."""
        return [value.serialize(options) for _, value in sorted(self.items())]

    @classmethod
    def deserialize(
        cls,
        data_sources: str,
        data_dicts_files: list[dict[str, Any]],
        options: Options,
        merge_options: MergeOptions,
    ) -> CoverageContainer:
        """Serialize the object."""
        covdata = CoverageContainer()
        for gcovr_file in data_dicts_files:
            if (
                filecov := FileCoverage.deserialize(
                    data_sources, gcovr_file, merge_options, options
                )
            ) is not None:
                covdata.insert_file_coverage(
                    filecov,
                    merge_options,
                )

        return covdata

    def merge_lines(self, options: Options) -> None:
        """Merge line coverage for same line number. Remove the function information on merged lines."""
        for filecov in self.values():
            filecov.merge_lines(
                is_file_excluded(
                    "trace",
                    filecov.filename,
                    options.trace_include_filter,
                    options.trace_exclude_filter,
                )
            )

    def merge(self, other: CoverageContainer, options: MergeOptions) -> None:
        """
        Merge CoverageContainer information and clear directory statistics.

        Do not use 'other' objects afterwards!
        """
        self.directories.clear()
        other.directories.clear()
        self.data.merge(other.data, options)

    def insert_file_coverage(
        self, filecov: FileCoverage, options: MergeOptions
    ) -> FileCoverage:
        """Add a file coverage item."""
        self.directories.clear()
        key = filecov.filename
        if key in self.data:
            self.data[key].merge(filecov, options)
        else:
            self.data[key] = filecov

        return filecov

    @property
    def stats(self) -> SummarizedStats:
        """Create a coverage statistic from a coverage data object."""
        stats = SummarizedStats.new_empty()
        for filecov in self.values():
            stats += filecov.stats
        return stats

    @staticmethod
    def _get_dirname(filename: str) -> str | None:
        """Get the directory name with a trailing path separator.

        >>> import os
        >>> CoverageContainer._get_dirname("bar/foobar.cpp".replace("/", os.sep)).replace(os.sep, "/")
        'bar/'
        >>> CoverageContainer._get_dirname("/foo/bar/A/B.cpp".replace("/", os.sep)).replace(os.sep, "/")
        '/foo/bar/A/'
        >>> CoverageContainer._get_dirname(os.sep) is None
        True
        """
        if filename == os.sep:
            return None
        return str(os.path.dirname(filename.rstrip(os.sep))) + os.sep

    def populate_directories(
        self, sorted_keys: Iterable[str], root_filter: re.Pattern[str]
    ) -> None:
        r"""Populate the list of directories and add accumulated stats.

        This function will accumulate statistics such that every directory
        above it will know the statistics associated with all files deep within a
        directory structure.

        Args:
            sorted_keys: The sorted keys for covdata
            root_filter: Information about the filter used with the root directory
        """

        # Get the directory coverage
        subdirs = dict[str, CoverageContainerDirectory]()
        for key in sorted_keys:
            filecov = self[key]
            dircov: CoverageContainerDirectory | None = None
            dirname: str | None = (
                os.path.dirname(filecov.filename)
                .replace("\\", os.sep)
                .replace("/", os.sep)
                .rstrip(os.sep)
            ) + os.sep
            while dirname is not None and root_filter.search(dirname + os.sep):
                if dirname not in subdirs:
                    subdirs[dirname] = CoverageContainerDirectory(dirname)
                if dircov is None:
                    subdirs[dirname][filecov.filename] = filecov
                else:
                    subdirs[dirname].data[dircov.filename] = dircov
                    subdirs[dircov.filename].parent_dirname = dirname
                subdirs[dirname].stats += filecov.stats
                dircov = subdirs[dirname]
                dirname = CoverageContainer._get_dirname(dirname)

        # Replace directories where only one sub container is available
        # with the content this sub container
        LOGGER.debug(
            "Replace directories with only one sub element with the content of this."
        )
        subdirs_to_remove = set()
        for dirname, covdata_dir in subdirs.items():
            # There is exact one element, replace current element with referenced element
            if len(covdata_dir) == 1:
                # Get the orphan item
                orphan_key, orphan_value = next(iter(covdata_dir.items()))
                # The only child is a File object
                if isinstance(orphan_value, FileCoverage):
                    # Replace the reference to ourself with our content
                    if covdata_dir.parent_dirname is not None:
                        LOGGER.debug(
                            "Move %s to %s.",
                            orphan_key,
                            covdata_dir.parent_dirname,
                        )
                        parent_covdata_dir = subdirs[covdata_dir.parent_dirname]
                        parent_covdata_dir[orphan_key] = orphan_value
                        del parent_covdata_dir[dirname]
                        subdirs_to_remove.add(dirname)
                else:
                    LOGGER.debug(
                        "Move content of %s to %s.",
                        orphan_value.dirname,
                        dirname,
                    )
                    # Replace the children with the orphan ones
                    covdata_dir.data = orphan_value.data
                    # Change the parent key of each new child element
                    for new_child_value in covdata_dir.values():
                        if isinstance(new_child_value, CoverageContainerDirectory):
                            new_child_value.parent_dirname = dirname
                    # Mark the key for removal.
                    subdirs_to_remove.add(orphan_key)

        for dirname in subdirs_to_remove:
            del subdirs[dirname]

        self.directories = list(subdirs.values())


class CoverageContainerDirectory(ContainerBase):
    """Represent coverage information about a directory."""

    __slots__ = "dirname", "parent_dirname", "data", "stats"

    def __init__(self, dirname: str) -> None:
        self.dirname: str = dirname
        self.parent_dirname: str | None = None
        self.data = CoverageDict[str, FileCoverage | CoverageContainerDirectory]()
        self.stats: SummarizedStats = SummarizedStats.new_empty()

    def __setitem__(
        self, key: str, item: FileCoverage | CoverageContainerDirectory
    ) -> None:
        self.data[key] = item

    def __getitem__(self, key: str) -> FileCoverage | CoverageContainerDirectory:
        return self.data[key]

    def __delitem__(self, key: str) -> None:
        del self.data[key]

    def __len__(self) -> int:
        return len(self.data)

    def values(self) -> ValuesView[FileCoverage | CoverageContainerDirectory]:
        """Get the file coverage data objects."""
        return self.data.values()

    def items(self) -> ItemsView[str, FileCoverage | CoverageContainerDirectory]:
        """Get the file coverage data items."""
        return self.data.items()

    def merge(self, other: CoverageContainerDirectory, options: MergeOptions) -> None:
        """
        Merge CoverageContainerDirectory information and clear directory statistics.

        Do not use 'other' objects afterwards!
        """
        self.data.merge(other.data, options)

    @property
    def filename(self) -> str:
        """Helpful function for when we use this DirectoryCoverage in a union with FileCoverage"""
        return self.dirname

    def line_coverage(self) -> CoverageStat:
        """A simple wrapper function necessary for sort_coverage()."""
        return self.stats.line

    def branch_coverage(self) -> CoverageStat:
        """A simple wrapper function necessary for sort_coverage()."""
        return self.stats.branch

    def decision_coverage(self) -> DecisionCoverageStat:
        """A simple wrapper function necessary for sort_coverage()."""
        return self.stats.decision



# -*- coding:utf-8 -*-

#  ************************** Copyrights and license ***************************
#
# This file is part of gcovr 8.6, a parsing and reporting tool for gcov.
# https://gcovr.com/en/8.6
#
# _____________________________________________________________________________
#
# Copyright (c) 2013-2026 the gcovr authors
# Copyright (c) 2013 Sandia Corporation.
# Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
# the U.S. Government retains certain rights in this software.
#
# This software is distributed under the 3-clause BSD License.
# For more information, see the README.rst file.
#
# ****************************************************************************

from lxml import etree  # nosec # We only write XML files

from ...data_model.container import CoverageContainer
from ...data_model.stats import CoverageStat, DecisionCoverageStat
from ...logging import LOGGER
from ...options import Options
from ...utils import write_xml_output


def write_report(
    covdata: CoverageContainer, output_file: str, options: Options
) -> None:
    """produce an XML report in the SonarQube generic coverage format"""

    if not any(
        filter(lambda filecov: filecov.condition_coverage().total > 0, covdata.values())  # type: ignore [arg-type]
    ) and (options.sonarqube_metric == "condition"):
        LOGGER.warning("No condition coverage data found.")

    root_elem = etree.Element("coverage")
    root_elem.set("version", "1")

    for _, filecov in sorted(covdata.items()):
        filename = filecov.presentable_filename(options.root_filter)

        file_node = etree.Element("file")
        file_node.set("path", filename)

        for linecov in filecov.linecov(sort=True):
            if linecov.is_reportable:
                line_node = etree.Element("lineToCover")
                line_node.set("lineNumber", str(linecov.lineno))
                line_node.set("covered", "true" if linecov.is_covered else "false")

                if options.sonarqube_metric != "line":
                    stat: CoverageStat | DecisionCoverageStat | None = None
                    if (
                        options.sonarqube_metric == "branch"
                        and linecov.has_reportable_branches
                    ):
                        stat = linecov.branch_coverage()
                    elif (
                        options.sonarqube_metric == "condition"
                        and linecov.has_reportable_conditions
                    ):
                        stat = linecov.condition_coverage()
                    elif options.sonarqube_metric == "decision" and linecov.decision:
                        stat = linecov.decision_coverage()

                    if stat:
                        line_node.set("branchesToCover", str(stat.total))
                        line_node.set("coveredBranches", str(stat.covered))

                file_node.append(line_node)

        root_elem.append(file_node)

    write_xml_output(
        root_elem,
        pretty=options.sonarqube_pretty,
        filename=output_file,
        default_filename="sonarqube.xml",
    )
